import operator
from functools import reduce
from typing import Type

from pynamodb.attributes import Attribute
from pynamodb.expressions.operand import Path
from pynamodb.models import Model

from pynamodb_utils.exceptions import FilterError
from pynamodb_utils.parsers import OPERATORS_MAPPING
from pynamodb_utils.utils import get_nested_attribute


def create_model_filter(model: Type[Model], args, _operator=operator.and_, raise_exception=True):
    conditions_list = []

    for key, value in args.items():
        array = key.rsplit('__', 1)
        field_name = array[0]
        operator_name = array[1] if len(array) > 1 and array[1] != 'not' else ''
        if operator_name.replace('not_', '') not in OPERATORS_MAPPING:
            raise FilterError(
                message={key: [f'Operator {operator_name} does not exist.'
                               f' Choose some of available: {", ".join(OPERATORS_MAPPING.keys())}']},
                status_code=400
            )
        nested_attr = get_nested_attribute(model, field_name, raise_exception)
        if isinstance(nested_attr, (Attribute, Path)):
            if 'not_' in operator_name:
                operator_name = operator_name.replace('not_', '')
                operator_handler = OPERATORS_MAPPING.get(operator_name)
                conditions_list.append(~operator_handler(model, field_name, nested_attr, value))
            else:
                operator_handler = OPERATORS_MAPPING.get(operator_name)
                conditions_list.append(operator_handler(model, field_name, nested_attr, value))

    if not conditions_list:
        return None
    else:
        return reduce(_operator, conditions_list)



from enum import Enum

import six
from pynamodb.attributes import MapAttribute, NumberAttribute, UnicodeAttribute


class DynamicMapAttribute(MapAttribute):
    element_type = None

    def __init__(self, *args, of=None, **kwargs):
        if of:
            if not issubclass(of, MapAttribute):
                raise ValueError("'of' must be subclass of MapAttribute")
            self.element_type = of
        super(DynamicMapAttribute, self).__init__(*args, **kwargs)

    def _set_attributes(self, **attributes):
        """
        Sets the attributes for this object
        """
        for attr_name, attr_value in six.iteritems(attributes):
            setattr(self, attr_name, attr_value)

    def deserialize(self, values):
        """
        Decode from map of AttributeValue types.
        """
        if not self.element_type:
            return super(DynamicMapAttribute, self).deserialize(values)

        class_for_deserialize = self.element_type()
        return {
            k: class_for_deserialize.deserialize(attr_value)
            for k, v in values.items()
            for _, attr_value in v.items()
        }

    @classmethod
    def is_raw(cls):
        return cls == DynamicMapAttribute

    def keys(self):
        return self.as_dict().keys()

    def items(self):
        return self.as_dict().items()

    def __repr__(self):
        return self.as_dict()

    def __str__(self):
        return str(self.__class__)


class EnumNumberAttribute(NumberAttribute):
    def __init__(self, hash_key=False, range_key=False, null=None, default=None, attr_name=None, enum=None):
        if isinstance(enum, Enum):
            raise ValueError('enum must be Enum class')
        self.enum = enum
        super().__init__(hash_key=hash_key, range_key=range_key, default=default, null=null, attr_name=attr_name)

    def serialize(self, value):
        if isinstance(value, self.enum):
            return value.value
        elif isinstance(value, str):
            if value in self.enum.__members__.keys():
                return getattr(self.enum, value).value
        raise ValueError(
            f'Value Error: {value} must be in {", ".join([item for item in self.enum.__members__.keys()])}')

    def deserialize(self, value):
        return self.enum(value).name


class EnumUnicodeAttribute(UnicodeAttribute):
    def __init__(self, hash_key=False, range_key=False, null=None, default=None, attr_name=None, enum=None):
        if isinstance(enum, Enum):
            raise ValueError('enum must be Enum class')
        self.enum = enum
        super().__init__(hash_key=hash_key, range_key=range_key, default=default, null=null, attr_name=attr_name)

    def serialize(self, value):
        if isinstance(value, self.enum):
            return value.value
        elif isinstance(value, str):
            if value in self.enum.__members__.keys():
                return getattr(self.enum, value).value
        raise ValueError(
            f'Value Error: {value} must be in {", ".join([item for item in self.enum.__members__.keys()])}')

    def deserialize(self, value):
        return self.enum(value).name


EnumAttribute = EnumNumberAttribute



class FilterError(Exception):
    message = ""
    status_code = 500

    def __init__(self, message=None, status_code=None):
        if message is not None:
            self.message = message
        if status_code is not None:
            self.status_code = status_code
        super().__init__(self.message, self.status_code)



from datetime import datetime
from typing import Dict, List, Union

from pynamodb import attributes
from pynamodb.expressions.operand import Path

from pynamodb_utils.attributes import EnumNumberAttribute, EnumUnicodeAttribute
from pynamodb_utils.exceptions import FilterError


def parse_string_to_datetime(value: str, field_name: str, *args):
    for fmt in DATETIME_FORMATS:
        try:
            return datetime.strptime(value, fmt)
        except ValueError:
            pass
    raise FilterError(message={field_name: [
        f"{value} is not valid type of {field_name}. Supported formats are {', '.join(DATETIME_FORMATS)}"]})


DATETIME_FORMATS = [
    '%Y-%m-%dT%H:%M:%S.%f+00:00',
    '%Y-%m-%dT%H:%M:%S.%f',
    '%Y-%m-%dT%H:%M:%S',
    '%Y-%m-%dT%H:%M',
    '%Y-%m-%d %H:%M:%S.%f+00:00',
    '%Y-%m-%d %H:%M:%S.%f',
    '%Y-%m-%d %H:%M:%S',
    '%Y-%m-%d %H:%M',
    '%Y-%m-%d'
]

NoneType = type(None)


def default_parser(value, *args):
    return value


def default_list_parser(value: List, field_name: str, *args):
    if isinstance(value, (list, NoneType)):
        return value
    raise FilterError(message={field_name: [f"{value} is not valid type of {field_name}."]})


def default_dict_parser(value: Dict, field_name: str, *args):
    if isinstance(value, (dict, NoneType)):
        return value
    raise FilterError(message={field_name: [f"{value} is not valid type of {field_name}."]})


def default_bool_parser(value: bool, field_name: str, *args):
    if isinstance(value, (bool, NoneType)):
        return value
    raise FilterError(message={field_name: [f"{value} is not valid type of {field_name}."]})


def default_str_parser(value, field_name: str, *args):
    if isinstance(value, (str, NoneType)):
        return value
    raise FilterError(message={field_name: [f"{value} is not valid type of {field_name}."]})


def default_number_parser(value: Union[float, int], field_name: str, *args):
    if isinstance(value, (float, int, NoneType)):
        return value
    raise FilterError(message={field_name: [f"{value} is not valid type of {field_name}."]})


def default_enum_parser(value, field_name: str, model):
    if isinstance(value, NoneType):
        return None

    values = getattr(model, field_name).enum.__members__

    _value = set(value) if isinstance(value, list) else {value}

    if _value.issubset(set(values)):
        return value

    raise FilterError(message={field_name: [
        f"{', '.join(value) if isinstance(value, list) else value} is not member of {', '.join(values)}."]})


TYPE_MAPPING = {
    attributes.UTCDateTimeAttribute: parse_string_to_datetime,
    attributes.UnicodeAttribute: default_str_parser,
    Path: default_parser,
    attributes.ListAttribute: default_list_parser,
    attributes.JSONAttribute: default_dict_parser,
    attributes.MapAttribute: default_parser,
    attributes.BooleanAttribute: default_bool_parser,
    attributes.NumberAttribute: default_number_parser,
    EnumNumberAttribute: default_enum_parser,
    EnumUnicodeAttribute: default_enum_parser
}


def parse_value(model, field_name: str, value):
    attrs = field_name.split('.')
    if len(attrs) == 1:
        _type = type(getattr(model, field_name))
        return TYPE_MAPPING[_type](value, field_name, model)
    else:
        return TYPE_MAPPING[attributes.MapAttribute](value, field_name, model)


def get_equals_condition(model, field_name: str, attr, value):
    parsed_value = parse_value(model, field_name, value)
    if parsed_value is None:
        return attr.does_not_exist()
    return attr.__eq__(parsed_value)


def get_startswith_condition(model, field_name: str, attr, value):
    return attr.startswith(parse_value(model, field_name, value))


def get_gt_condition(model, field_name: str, attr, value):
    return attr.__gt__(parse_value(model, field_name, value))


def get_lt_condition(model, field_name: str, attr, value):
    return attr.__lt__(parse_value(model, field_name, value))


def get_gte_condition(model, field_name: str, attr, value):
    return attr.__ge__(parse_value(model, field_name, value))


def get_lte_condition(model, field_name: str, attr, value):
    return attr.__le__(parse_value(model, field_name, value))


def get_contains_condition(model, field_name: str, attr, value):
    parsed_value = parse_value(model, field_name, value)
    if isinstance(parsed_value, list):
        return attr.contains(*parsed_value)
    return attr.contains(parsed_value)


def get_is_in_condition(model, field_name: str, attr, value):
    parsed_value = parse_value(model, field_name, value)
    if isinstance(parsed_value, list):
        return attr.is_in(*parsed_value)
    return attr.is_in(parsed_value)


OPERATORS_MAPPING = {
    '': get_equals_condition,
    'equals': get_equals_condition,
    'contains': get_contains_condition,
    'startswith': get_startswith_condition,
    'gt': get_gt_condition,
    'lt': get_lt_condition,
    'gte': get_gte_condition,
    'lte': get_lte_condition,
    'is_in': get_is_in_condition,
}



from datetime import timezone

from marshmallow.exceptions import ValidationError
from pynamodb.attributes import UTCDateTimeAttribute
from pynamodb.models import Model

from pynamodb_utils.exceptions import FilterError
from pynamodb_utils.query_serializer import QuerySerializer
from pynamodb_utils.utils import get_timestamp, parse_attrs_to_dict


class JSONQueryModel(Model):

    class Meta:
        abstract = True

    @classmethod
    def get_conditions_from_json(cls, query: dict):
        try:
            return QuerySerializer().load(model=cls, data=query)
        except ValidationError as e:
            raise FilterError(message=e.messages, status_code=400)


class AsDictModel(Model):

    class Meta:
        abstract = True

    def as_dict(self):
        return parse_attrs_to_dict(self)


class TimestampedModel(Model):
    created_at = UTCDateTimeAttribute(default=get_timestamp)
    updated_at = UTCDateTimeAttribute(default=get_timestamp)
    deleted_at = UTCDateTimeAttribute(null=True)

    class Meta:
        abstract = True

    def save(self, condition=None):
        tz_info = getattr(self.Meta, "TZINFO", None)
        self.created_at = self.created_at.astimezone(tz=tz_info or timezone.utc)
        self.updated_at = get_timestamp(tzinfo=tz_info)
        super().save(condition=condition)

    def save_without_timestamp_update(self, condition=None):
        super().save(condition=condition)

    def soft_delete(self, condition=None):
        tz_info = getattr(self.Meta, "TZINFO", None)
        self.deleted_at = get_timestamp(tz_info)
        super().save(condition=condition)



from datetime import datetime, timezone

from pynamodb.attributes import MapAttribute

from pynamodb_utils.attributes import DynamicMapAttribute
from pynamodb_utils.exceptions import FilterError


def parse_attr(attr):
    if isinstance(attr, DynamicMapAttribute):
        return attr.as_dict()
    elif isinstance(attr, MapAttribute):
        return parse_attrs_to_dict(attr)
    elif isinstance(attr, datetime):
        return str(attr)
    return attr


def parse_attrs_to_dict(obj):
    return {k: parse_attr(getattr(obj, k, None)) for k in obj.get_attributes().keys()}


def get_nested_attribute(model, attr_string, raise_exception=True):
    attrs = attr_string.split('.')
    result = model
    for attr in attrs:
        if isinstance(result, DynamicMapAttribute):
            result = result[attr]
        elif hasattr(result, attr):
            result = getattr(result, attr)
        elif raise_exception:
            raise FilterError(
                message={attr_string: [f'Parameter {attr} does not exist.'
                                       f' Choose some of available: {", ".join(result.get_attributes())}']},
                status_code=400
            )
        else:
            return None
    return result


def get_timestamp(tzinfo=None):
    return datetime.now(tzinfo or timezone.utc)



import operator
import typing
from functools import reduce

from marshmallow import INCLUDE, Schema, fields, post_load, types
from marshmallow.exceptions import ValidationError

from pynamodb_utils.filters import FilterError, create_model_filter


class QuerySerializer(Schema):
    Or = fields.List(fields.Dict, required=False, allow_none=True)
    And = fields.List(fields.Dict, required=False, allow_none=True)

    class Meta:
        unknown = INCLUDE

    def __init__(
            self,
            *,
            only: types.StrSequenceOrSet = None,
            exclude: types.StrSequenceOrSet = (),
            many: bool = False,
            context: typing.Dict = None,
            load_only: types.StrSequenceOrSet = (),
            dump_only: types.StrSequenceOrSet = (),
            partial: typing.Union[bool, types.StrSequenceOrSet] = False,
            unknown: str = None,
            model=None,
            raise_exception=True
    ):
        super().__init__(only=only, exclude=exclude, many=many, context=context, load_only=load_only,
                         dump_only=dump_only, partial=partial, unknown=unknown)
        self.model = model
        self.raise_exception = raise_exception

    def load(
            self,
            data: typing.Mapping,
            *,
            many: bool = None,
            partial: typing.Union[bool, types.StrSequenceOrSet] = None,
            unknown: str = None,
            model=None,
            raise_exception=True
    ):
        if model is not None:
            self.model = model
            self.raise_exception = raise_exception
        return super().load(data=data, many=many, partial=partial, unknown=unknown)

    @post_load(pass_original=True)
    def create_query(self, data, original_data, **kwargs):
        try:
            conditions = []
            or_conditions = original_data.pop('Or', [])
            and_conditions = original_data.pop('And', [])

            query = create_model_filter(model=self.model, args=original_data, raise_exception=self.raise_exception)
            or_conditions = [create_model_filter(model=self.model, args=item, raise_exception=self.raise_exception) for
                             item in or_conditions]
            and_conditions = [create_model_filter(model=self.model, args=item, raise_exception=self.raise_exception) for
                              item in and_conditions]

            if or_conditions:
                conditions.append(reduce(operator.or_, or_conditions))
            if and_conditions:
                conditions.append(reduce(operator.and_, and_conditions))

            if query is not None:
                conditions.append(query)

            return reduce(operator.and_, conditions) if conditions else None
        except ValueError as e:
            raise ValidationError(message=str(e))
        except FilterError as e:
            raise ValidationError(
                message={
                    "Query": e.message
                }
            )
